/**
* Count the number of itemsets at frequency theta and theta-epsilon
*
* Copyright 2015 Matteo Riondato <matteo@cs.brown.edu>
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

#include <cerrno>
#include <cstdio>
#include <cstdlib>
#include <iostream>

#include <unistd.h>
extern char *optarg;
extern int optind;

#include "util.h"

using namespace std;

void usage(char *binary_name) {
	cerr << binary_name
		<< ": count the FIs wrt theta and to theta-epsilon" << endl;
	cerr << "USAGE: "
		<< binary_name << "[-h] epsilon theta exact_results_file" << endl;
	cerr << "\t-h: print this help message and exit" << endl;
}

int main(int argc, char **argv) {
	int opt;
	while ((opt = getopt(argc, argv, "h")) != -1) {
		switch (opt) {
		case 'h':
			usage(argv[0]);
			return EXIT_SUCCESS;
			break;
		}
	}
	if (optind != argc - 3) {
		cerr << ERROR_HEADER << "wrong number of arguments" << endl;
		return EXIT_FAILURE;
	}
	const double epsilon = strtod(argv[argc - 3], NULL);
	if (errno == ERANGE || epsilon <= 0.0 || epsilon >= 1.0) {
		cerr << ERROR_HEADER
			<< "epsilon must be a number greater than 0.0 and smaller than 1"
			<< endl;
		return EXIT_FAILURE;
	}
	const double theta = strtod(argv[argc - 2], NULL);
	if (errno == ERANGE || theta <= 0.0 || theta >= 1.0) {
		cerr << ERROR_HEADER
			<< "theta must be a number greater than 0.0 and smaller than 1"
			<< endl;
		return EXIT_FAILURE;
	}
	FILE *exact_FILE = fopen(argv[argc - 1], "r");
	if (exact_FILE == NULL) {
		perror("Error opening exact_results_file");
		return errno;
	}

	char line[TRANSACTION_LINE_MAXLEN];
	get_next_line(line, TRANSACTION_LINE_MAXLEN, exact_FILE);
	const unsigned int size = (unsigned int) atoi(line + 1);
	double previous_freq = 2.0;
	unsigned int res_count = 0u;
	unsigned int extended_res_count = 0u;
	while (get_next_line(line, TRANSACTION_LINE_MAXLEN, exact_FILE)) {
		char *parenthesis = strchr(line, '(');
		*(parenthesis - 1) = '\0';
		const double support = strtod(parenthesis + 1, NULL);
		const double freq = support / size;
		if (freq > previous_freq) {
			cerr << ERROR_HEADER << "results must be sorted" << endl;
			return EXIT_FAILURE;
		}
		if (freq >= theta) {
			++res_count;
			++extended_res_count;
		} else if (freq >= theta - epsilon) {
			++extended_res_count;
		} else {
			break;
		}
	}
	cout << "file,theta,epsilon,extended,exact,diff" << endl;
	cout << argv[argc - 1] << "," << theta << "," << epsilon << "," <<
		extended_res_count << "," << res_count << "," <<
		extended_res_count - res_count << endl;
	return 0;
}
